/*
 *  Copyright (C) 1998,9 by Marco G"otze.
 *
 *  This code is part of the wmpinboard source package, which is
 *  distributed under the terms of the GNU GPL2.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "wmpinboard.h"
#include "xmisc.h"
#include "notes.h"

#include "pinboard.xpm"
#include "pinboard.xbm"

static const char *rc_file_name = "/.wmpinboarddata";

#define NUM_FONTS 3
static const struct {  /* shortcuts for known fonts */
  char name[8];
  char desc[40];
  char font[60];
} fonts[NUM_FONTS] = {
  { "Default", "default font (Latin1 (ISO8859-1))",
    "-*-fixed-*--10-*-iso8859-1" },
  { "Latin2",  "Latin2 (ISO8859-2)",
    "-*-fixed-*--10-*-iso8859-2" },
  { "Russian", "cyrillic font (KOI8)",
    "-*-fixed-*--10-*-koi8-r" }
};
static const char *default_font = fonts[0].font;

static struct {  /* options and parameters */
  char *name;               /* the program's file name (argv[0]) */
  char *display;            /* alternate X display to connect to */
  int click_to_focus;       /* true if keyboard focus requires a click */
  int no_timeout;           /* true if timeout should be disabled */
  char *user_font;          /* user-specified font */
  char font_to_remember[STRING_BUF_SIZE];  /* font descriptor to remember */
} opts = { 0, 0, 0,
#if TIMEOUT > 0
  0,
#endif
  0, "" };

data_t ndata[MAX_NOTES];  /* this holds all the notes */
int notes_count = 0;

Display     *display;
Window      Root;
int         screen;
int         x_fd;
int         d_depth;
XSizeHints  mysizehints;
XWMHints    mywmhints;
Pixel       back_pix, fore_pix;
char        *Geometry = "";
Window      iconwin, win;
GC          NormalGC;
XpmIcon     app;
Pixmap      pixmask;

XImage *palette, *board_img, *note_img, *char_img = 0;

                              /*******************/
/****************************** IMPLEMENTATIONS ******************************/
                            /*******************/
/*
 * writes all notes to or loads them from a text file
 */
void
notes_io(int save)
{
  static char *ext = ".new";
  static char *header = "WMPB0"; /* number indicates the file format revision */
  static char s[STRING_BUF_SIZE];
  char t[STRING_BUF_SIZE];
  FILE *file;

  s[STRING_BUF_SIZE-1] = '\0';
  strncpy(s, getenv("HOME"), STRING_BUF_SIZE-1);
  if (STRING_BUF_SIZE-strlen(s)-1 < strlen(rc_file_name+strlen(ext))) {
    fprintf(stderr, "Fatal error: buffer too small in notes_io().\n");
    exit(1);
  }
  strcat(s, rc_file_name);
  if (save) {  /* save to temporary file first, later rename it */
    strcpy(t, s);
    strcat(s, ext);
  }

  file = fopen(s, save ? "wb" : "rb");
  if (!file) {
    if (save) {
      fprintf(stderr, "Fatal error: couldn't open rc file for writing.\n");
      exit(1);
    } else
      return;  /* just don't read in anything if opening the file failed */
  }
  if (save) {
    fwrite(header, 5, 1, file);
    fputs(opts.font_to_remember, file);
    fputc('\n', file);
    fwrite(ndata, sizeof(data_t), notes_count, file);
    fclose(file);

    strcpy(s, t);
    strcat(s, ext);
    if (rename(s, t) == -1) {
      fprintf(stderr, "Fatal error: failed to rename file `%s' to `%s'.\n", t,
        s);
      exit(1);
    }
  } else {  /* load */
    if (fread(s, 5, 1, file)) {
      if (strncmp(s, header, 4)) {
        fprintf(stderr, "Fatal error: corrupt data file encountered.\n"
          "Delete `~%s' to start over.\n", rc_file_name);
        fclose(file);
        exit(1);
      }
      /* no file format revision check necessary yet */
      fgets(opts.font_to_remember, STRING_BUF_SIZE, file);
      if (opts.font_to_remember[strlen(opts.font_to_remember)-1] == '\n')
        opts.font_to_remember[strlen(opts.font_to_remember)-1] = '\0';
      notes_count = fread(ndata, sizeof(data_t), MAX_NOTES, file);
    }
    fclose(file);
  }
}

/*
 * initializes the application's X window
 */
void
init()
{
  unsigned int borderwidth = 1;
  XClassHint classHint;
  char *display_name = opts.display;
  char *wname = opts.name;
  XTextProperty name;
  XGCValues gcv;
  unsigned long gcm;
  int i, dummy = 0;

  if (!(display = XOpenDisplay(display_name))) {
    fprintf(stderr, "Fatal error: Can't open display %s.\n", 
      XDisplayName(display_name));
    exit(1);
  }

  screen  = DefaultScreen(display);
  Root    = RootWindow(display, screen);
  d_depth = DefaultDepth(display, screen);
  x_fd    = XConnectionNumber(display);

  get_xpm(&app, pinboard_xpm);  /* convert app back image */

  mysizehints.flags = USSize | USPosition;
  mysizehints.x = 0;
  mysizehints.y = 0;

  back_pix = lookup_color("white");
  fore_pix = lookup_color("black");

  XWMGeometry(display, screen, Geometry, 0, borderwidth, &mysizehints,
    &mysizehints.x, &mysizehints.y, &mysizehints.width, &mysizehints.height,
    &dummy);

  mysizehints.width = 64;
  mysizehints.height = 64;
    
  win = XCreateSimpleWindow(display, Root, mysizehints.x, mysizehints.y,
    mysizehints.width, mysizehints.height, borderwidth, fore_pix, back_pix);
  
  iconwin = XCreateSimpleWindow(display, win, mysizehints.x, mysizehints.y,
    mysizehints.width, mysizehints.height, borderwidth, fore_pix, back_pix);

  XSetWMNormalHints(display, win, &mysizehints);

  mywmhints.initial_state = WithdrawnState;
  mywmhints.icon_window = iconwin;
  mywmhints.icon_x = mysizehints.x;
  mywmhints.icon_y = mysizehints.y;
  mywmhints.window_group = win;
  mywmhints.flags = StateHint | IconWindowHint | WindowGroupHint;
  XSetWMHints(display, win, &mywmhints);

  classHint.res_name = wname;
  classHint.res_class = wname;
  XSetClassHint(display, win, &classHint);

  XSelectInput(display, win, ButtonPressMask | ExposureMask |
    ButtonReleaseMask | PointerMotionMask | StructureNotifyMask |
    KeyPressMask | EnterWindowMask | LeaveWindowMask | FocusChangeMask);
  XSelectInput(display, iconwin, ButtonPressMask | ExposureMask |
    ButtonReleaseMask | PointerMotionMask | StructureNotifyMask |
    KeyPressMask | EnterWindowMask | LeaveWindowMask | FocusChangeMask);

  if (!XStringListToTextProperty(&wname, 1, &name)) {
    fprintf(stderr, "Fatal error: Can't allocate window name.\n");
    exit(1);
  }

  XSetWMName(display, win, &name);

  /* create GC... */
  gcm = GCForeground | GCBackground | GCGraphicsExposures;
  gcv.foreground = fore_pix;
  gcv.background = back_pix;
  gcv.graphics_exposures = 0;
  NormalGC = XCreateGC(display, Root, gcm, &gcv);

  pixmask = XCreateBitmapFromData(display, win, pinboard_bits, pinboard_width,
    pinboard_height);
  set_mask(1);
  XMapWindow(display, win);
}

/*
 * prepares a set of character images for use with in edit mode
 */
void
prepare_font()
{
  XGCValues gcv;
  unsigned long gcm;
  GC fontGC;
  XFontStruct *font = 0;
  Pixmap char_pixmap;
  int i;

  /*
   * tries to load font name, dies upon an error if die is true
   */
  int
  try_font(const char *name, int die)
  {
    if (font) return 1;
    if (!name || !*name) return 0;
    font = XLoadQueryFont(display, name);
    if (font) {
      if ((font->max_bounds.rbearing - font->min_bounds.lbearing != 6 &&
        font->max_bounds.rbearing - font->min_bounds.lbearing != 7) ||
        font->max_bounds.ascent + font->max_bounds.descent != 10)
      {
        fprintf(stderr, "The font `%s'\n"
          "  doesn't have a fixed character cell size of 6x10.\n", name);
        XFreeFont(display, font);
        font = 0;
      } else {
        if (!notes_count)
          strncpy(opts.font_to_remember, name, STRING_BUF_SIZE);
        return 1;
      }
    } else
      fprintf(stderr, "The font `%s' doesn't exist.\n", name);
    if (die) {
      fprintf(stderr, "No alternatives left, aborting.\n");
      exit(1);
    } else
      fprintf(stderr, "Trying to revert...\n");
    return 0;
  }

  if (!notes_count)
    try_font(opts.user_font, 0);
  try_font(opts.font_to_remember, 0);
  try_font(default_font, 1);

  char_pixmap = XCreatePixmap(display, app.pixmap, 255*6+1, 10,
    DefaultDepth(display, DefaultScreen(display)));

  gcm = GCForeground | GCBackground | GCGraphicsExposures | GCFont;
  gcv.foreground = get_color(C_INNER);
  gcv.background = get_color(C_OUTER);
  gcv.graphics_exposures = 0;
  gcv.font = font->fid;
  fontGC = XCreateGC(display, char_pixmap, gcm, &gcv);
  
  for (i = 1; i < 256; i++) {
    char ch = i;
    XDrawImageString(display, char_pixmap, fontGC, (i-1)*6, font->ascent,
      &ch, 1);
  }
  XFreeGC(display, fontGC);
  XFreeFont(display, font);

  if (char_img)
    XDestroyImage(char_img);
  char_img = XGetImage(display, char_pixmap, 0, 0, 255*6+1, 10, ~0, ZPixmap);
  XFreePixmap(display, char_pixmap);
  XPutPixel(char_img, 255*6, 0, get_color(C_INNER));
  XPutPixel(char_img, 255*6, 1, get_color(C_OUTER));
}

/*
 * (un)sets the keyboard input focus to wmpinboard's window
 */  
void 
set_kbfocus(int get_it)
{
  XWindowAttributes w_attr;

  if (get_it) {
    XGetWindowAttributes(display, iconwin, &w_attr);
    if ((w_attr.map_state & IsViewable) == IsViewable)
      XSetInputFocus(display, iconwin, RevertToPointerRoot, CurrentTime);
    else
      XSetInputFocus(display, win, RevertToPointerRoot, CurrentTime);
  } else if (!opts.click_to_focus)
    XSetInputFocus(display, PointerRoot, RevertToNone, CurrentTime);
}

/*
 * does some preparations for note edit mode
 */
void
init_edit_mode(int note)
{
  if (XGetPixel(char_img, 255*6, 1) == get_text_color(ndata[note].col))
    replace_color(char_img, 0, 0, 255*6+1, 10, XGetPixel(char_img, 255*6, 1),
      get_text_color(C_OUTER));
  replace_color(char_img, 0, 0, 255*6, 10, XGetPixel(char_img, 255*6, 0),
    get_text_color(ndata[note].col));
  replace_color(char_img, 0, 0, 255*6, 10, XGetPixel(char_img, 255*6, 1),
    get_color(ndata[note].col));
  XPutPixel(char_img, 255*6, 0, get_text_color(ndata[note].col));
  XPutPixel(char_img, 255*6, 1, get_color(ndata[note].col));
  set_mask(0);
  color_note(note);
  update_display(note_img, 0, 0);
  print_text(note);
}

/*
 * parse the list of command line arguments and initializes the application's
 * opts structure accordingly
 */
void
parse_argv(int argc, char **argv)
{
  int i, j;

  /*
   * print help screen
   */
  void
  help(void)
  {
    puts("wmpinboard v" VERSION "\n");
    puts("Copyright (C) 1998,9 by Marco G\"otze, "
      "<mailto:gomar@mindless.com>.");
    puts("This program is distributed under the terms of the GNU GPL2.\n");
    puts("usage: wmpinboard [-display DISPLAY] [-c] [-f font] [-t] [--help]");
    puts("  -display  use the specified X display");
    puts("  -c        emulate click-based keyboard focus");
    puts("  -f FONT   use the specified font; FONT can be one of the "
      "following:");
    for (i = 0; i < NUM_FONTS; i++)
      printf("              %-8s  %s\n", fonts[i].name, fonts[i].desc);
    puts("            or a complete X descriptor of a fixed size 6x10 font");
#if TIMEOUT > 0
    puts("  -t        disable timeout (automatic edit mode termination)");
#endif
    puts("  --help    print this help");
    exit(1);
  }

  opts.name = argv[0];
  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], "-display")) {  /* user-specified display */
      if (i++ == argc-1)  /* parameter missing */
        help();
      opts.display = argv[i];
    } else if (!strcmp(argv[i], "-c"))
      opts.click_to_focus = 1;
    else if (!strcmp(argv[i], "-f")) {  /* user-specified font */
      if (i++ == argc-1)  /* parameter missing */
        help();
      for (j = 0; j < NUM_FONTS; j++)
        if (!strcasecmp(argv[i], fonts[j].name)) {
          opts.user_font = (char*) fonts[j].font;
          break;
        }
      if (j >= NUM_FONTS) {  /* not a predefined font */
        if (strlen(argv[i]) >= STRING_BUF_SIZE) {
          fprintf(stderr, "Fatal error: specified font descriptor exceeds "
            "buffer size of %d.\n", STRING_BUF_SIZE);
          exit(1);
        }
        opts.user_font = argv[i];
      }
    }
#if TIMEOUT > 0
    else if (!strcmp(argv[i], "-t"))
      opts.no_timeout = 1;
#endif
    else
      help();
}
                                   /********/
/*********************************** MAIN ************************************/
                                 /********/
int
main(int argc, char **argv)
{
  XEvent event;
  KeySym ksym;
  int cur_note;  /* note currently being processed */
  int button = -1, dx, dy;  /* mouse-related stuff */
  int mode = M_NOOP;
  int cursor = 0, insert = 0;
  int i, j;
#if TIMEOUT > 0
  time_t idle;
#endif

  /*
   * in edit mode, moves the cursor to a new position and updates the display
   */
  void
  set_cursor(int new_pos)
  {
    unsigned long c = get_text_color(ndata[cur_note].col), d;
    int i, j;

    if (new_pos > 58)
      return;
    print_letter(cur_note, cursor);
    cursor = new_pos;
    if (insert) {
      d = get_color(ndata[cur_note].col);
      for (i = 2+6*(cursor%10); i < 2+6*(cursor%10 + 1); i++)
        for (j = 2+10*(cursor/10); j < 2+10*(cursor/10 + 1); j++)
          XPutPixel(note_img, i, j, XGetPixel(note_img, i, j) == c ? d : c);
    } else
      for (i = 0; i < 6; i++)
        XPutPixel(note_img, 2+6*(cursor%10)+i, 11+10*(cursor/10), c);
    update_display(note_img, 0, 0);
  }

  srand(time(0));
  umask(077);  /* users' notes needn't be world-readable */

  parse_argv(argc, argv);  /* evaluate command line parameters */
  init();                  /* initialize X window */
  XSetCommand(display, win, argv, argc);
  init_xlocale();          /* initialize input context */

  palette = XGetImage(display, app.pixmap, 144, 0, 2, 64, ~0, ZPixmap);
  note_img = XGetImage(display, app.pixmap, 0, 0, 128, 64, ~0, ZPixmap);
  XCopyArea(display, app.pixmap, app.pixmap, NormalGC, 0, 0, 64, 64,
    64, 0);  /* copy empty pinboard image */
  board_img = XGetImage(display, app.pixmap, 0, 0, 144, 64, ~0, ZPixmap);
  copy_image_area(board_img, board_img, 0, 0, 64, 0, 64, 64);

  notes_io(0);
  render_pinboard(-1);
  update_display(board_img, 0, 0);
  prepare_font();
  notes_io(1);  /* saves opts.font_to_remember */

  for(;;) {
    static XComposeStatus compose = { 0, 0 };
    static unsigned char ch[4];
    static int clicks_count = 0;

    while (XPending(display)) {
      XNextEvent(display, &event);
      switch (event.type) {
        case Expose:
          redraw_window();
          break;
        case ButtonPress:
          button = event.xbutton.button;
          if (mode == M_NOOP) {
            cur_note = selected_note(event.xbutton.x, event.xbutton.y);
            if (cur_note >= 0 && button > 1) {
              cur_note = raise_note(cur_note);
              dx = event.xbutton.x - ndata[cur_note].x;
              dy = event.xbutton.y - ndata[cur_note].y;
              render_pinboard(cur_note);
              color_notes(cur_note);
              pin_note(cur_note);
              update_display(board_img, 0, 0);
              mode = M_MOVE;
            }
          } else if (opts.click_to_focus && mode == M_EDIT && !clicks_count)
            set_kbfocus(1);
          break;
        case ButtonRelease:
          clicks_count++;
          if (event.xbutton.button == button) {
            if (mode == M_MOVE) {  /* note was being dragged */
              pin_note(cur_note);
              update_display(board_img, 0, 0);
              copy_image_area(board_img, board_img, 0, 0, 64, 0, 64, 64);
              notes_io(1);
              mode = M_NOOP;
            } else if (mode == M_NOOP && button == 1 && /* add new note? */
              event.xbutton.x >= 18 && event.xbutton.x <= 45 &&
              event.xbutton.y >=  2 && event.xbutton.y <= 12 &&
              selected_note(event.xbutton.x, event.xbutton.y) < 0)
            {
              if ((cur_note = add_note()) >= 0) {
                init_edit_mode(cur_note);
                set_cursor(0);
                set_kbfocus(1);
                mode = M_EDIT;
#if TIMEOUT > 0
                time(&idle);
#endif
              } else
                XBell(display, 0);
            } else if (mode == M_EDIT) { 
              if (event.xbutton.x + event.xbutton.y >= 115) {  /* leave? */
                if (button > 1 && !string_empty(ndata[cur_note].text, 0)) {
                  for (i = 0; i < 59; i++)
                    ndata[cur_note].text[i] = ' ';
                  print_text(cur_note);
                  set_cursor(0);
#if TIMEOUT > 0
                  time(&idle);
#endif
                  break;
                }
                if (string_empty(ndata[cur_note].text, 0))
                  remove_note(cur_note);
                set_mask(1);
                render_pinboard(-1);
                copy_image_area(board_img, board_img, 0, 0, 64, 0, 64, 64);
                update_display(board_img, 0, 0);
                notes_io(1);
                set_kbfocus(0);
                mode = M_NOOP;
              } else {
                if (!opts.click_to_focus || clicks_count > 1) {
                  i = ndata[cur_note].col;
                  if (event.xbutton.button == 1) {  /* next color */
                    if (++ndata[cur_note].col == C_NUM) ndata[cur_note].col = 0;
                  } else if (--ndata[cur_note].col < 0)  /* previous color */
                      ndata[cur_note].col = C_NUM-1;
                  init_edit_mode(cur_note);
                  set_cursor(cursor);
                }
#if TIMEOUT > 0
                time(&idle);
#endif
              }
            } else if (mode == M_NOOP && button == 1 && cur_note >= 0 &&
              cur_note == selected_note(event.xbutton.x, event.xbutton.y))
            {
              init_edit_mode(cur_note);
              set_cursor(0);
              set_kbfocus(1);
              mode = M_EDIT;
#if TIMEOUT > 0
              time(&idle);
#endif
            }
          }
          button = -1;
          break;
        case MotionNotify:
          if (mode == M_MOVE && button > 1) {
            ndata[cur_note].x = event.xbutton.x - dx;
            ndata[cur_note].x = ndata[cur_note].x < 6 ? 6 : ndata[cur_note].x;
            ndata[cur_note].x =
              ndata[cur_note].x > 42 ? 42 : ndata[cur_note].x;
            ndata[cur_note].y = event.xbutton.y - dy;
            ndata[cur_note].y = ndata[cur_note].y < 2 ? 2 : ndata[cur_note].y;
            ndata[cur_note].y =
              ndata[cur_note].y > 46 ? 46 : ndata[cur_note].y;
            pin_note(cur_note);
            update_display(board_img, 0, 0);
          }
          break;
        case EnterNotify:
          if (mode == M_EDIT && !opts.click_to_focus) set_kbfocus(1);
          clicks_count = 0;
          break;
        case LeaveNotify:
          if (mode == M_EDIT && !opts.click_to_focus) set_kbfocus(0);
          break;
        case KeyPress:
#if TIMEOUT > 0
          time(&idle);
#endif
          if (mode != M_EDIT || event.xkey.state & 0xdfec ||
            (InputContext && XFilterEvent(&event, win)))
          {
            XSendEvent(display, RootWindow(display, DefaultScreen(display)),
              True, KeyPressMask, &event);
            if (mode == M_EDIT) set_kbfocus(1);  /* make sure we keep focus */
            break;
          }
          j = 0;
          if (InputContext)
            j = XmbLookupString(InputContext, &event.xkey, ch, sizeof(ch),
              &ksym, 0);
          else
            j = XLookupString(&event.xkey, ch, sizeof(ch), &ksym, &compose);
          if (!j && (ksym & 0xff00) == 0x0600) {  /* cyrillic keysyms */
            ch[0] = (unsigned char) (ksym & 0x00ff);
            ch[1] = '\0';
            if (ch[0] >= 0xbf || ch[0] == 0xa3 || ch[0] == 0xb3)
              j = 1;  /* filter extended KOI8 characters */
          }
          switch (ksym) {
            case XK_Return: case XK_KP_Enter:
              if (cursor >= 50)
                break;
              if (insert) {
                shift_string(cur_note, 10*(cursor/10+1), 58, cursor%10, ' ');
                shift_string(cur_note, cursor, 58, 10-cursor%10, ' ');
              }
              set_cursor(10*(cursor/10+1));
              break;
            case XK_BackSpace: case 0xfd1a:
              if (!cursor)
                break;
              print_letter(cur_note, cursor--);
              ndata[cur_note].text[cursor] = ' ';
            case XK_Delete: case XK_KP_Delete:
              j = 1;  /* delete entire line if empty... */
              if (((ksym != XK_BackSpace && ksym != 0xfd1a) ||
                cursor%10 == 9) &&
                string_empty(&ndata[cur_note].text[10*(cursor/10)], 10))
              {
                cursor = 10*(cursor/10);
                j = 10;
              }
              for (i = cursor;
                i < (j == 1 && cursor/10 < 5 ? 10*(cursor/10+1)-j : 59-j); i++)
              {
                ndata[cur_note].text[i] = ndata[cur_note].text[i+j];
                ndata[cur_note].text[i+j] = ' ';
                print_letter(cur_note, i);
              }
              if (cursor%10 == 9 || cursor == 58)  /* exceptions to the loop */
                ndata[cur_note].text[cursor] = ' ';
              if (j > 1)
                print_text(cur_note);
              else
                print_letter(cur_note, i);
              set_cursor(cursor);
              break;
            case XK_Up: case XK_KP_Up:
              if ((event.xkey.state & 0x01) == 0x01) {  /* next note w/s/col */
                for (i = cur_note == notes_count-1 ? 0 : cur_note+1;
                  i != cur_note;)
                {
                   if (ndata[i].col == ndata[cur_note].col) break;
                   if (++i >= notes_count) i = 0;
                }
                if (i == cur_note)
                  break;
                if (string_empty(ndata[cur_note].text, 0))
                  remove_note(cur_note);
                notes_io(1);
                cur_note = i;
                init_edit_mode(cur_note);
                set_cursor(0);
              } else  /* move cursor */
                if (cursor > 9) set_cursor(cursor-10);
              break;
            case XK_Down: case XK_KP_Down:
              if ((event.xkey.state & 0x01) == 0x01) {  /* prev. note w/s/col */
                for (i = cur_note ? cur_note-1 : notes_count-1; i != cur_note;)
                {
                   if (ndata[i].col == ndata[cur_note].col) break;
                   if (--i < 0) i = notes_count-1;
                }
                if (i == cur_note)
                  break;
                if (string_empty(ndata[cur_note].text, 0))
                  remove_note(cur_note);
                notes_io(1);
                cur_note = i;
                init_edit_mode(cur_note);
                set_cursor(0);
              } else  /* move cursor */
                if (cursor < 49) set_cursor(cursor+10);
              break;
            case XK_Left: case XK_KP_Left:
              if ((event.xkey.state & 0x01) == 0x01) {  /* previous note */
                if (notes_count == 1)
                  break;
                if (string_empty(ndata[cur_note].text, 0))
                  remove_note(cur_note);
                notes_io(1);
                if (--cur_note < 0)
                  cur_note = notes_count-1;
                init_edit_mode(cur_note);
                set_cursor(0);
              } else  /* move cursor */
                if (cursor) set_cursor(cursor-1);
              break;
            case XK_Right: case XK_KP_Right:
              if ((event.xkey.state & 0x01) == 0x01) {  /* next note */
                if (notes_count == 1)
                  break;
                if (string_empty(ndata[cur_note].text, 0)) {
                  remove_note(cur_note);
                  cur_note = notes_count-1;
                }
                notes_io(1);
                if (++cur_note == notes_count)
                  cur_note = 0;
                init_edit_mode(cur_note);
                set_cursor(0);
              } else  /* move cursor */
                if (cursor < 58) set_cursor(cursor+1);
              break;
            case XK_Home: case XK_KP_Home:
              set_cursor(10*(cursor/10));
              break;
            case XK_End: case XK_KP_End:
              for (i = cursor < 50 ? 10*(cursor/10)+9 : 58;
                i > 10*(cursor/10) && ndata[cur_note].text[i-1] == ' '; i--) {}
              set_cursor(i);
              break;
            case XK_Prior: case XK_KP_Prior:  /* page up */
              set_cursor(0);
              break;
            case XK_Next: case XK_KP_Next:  /* page down */
              set_cursor(58);
              break;
            case XK_Insert: case XK_KP_Insert:  /* insert */
              insert = !insert;
              set_cursor(cursor);
              break;
            default:
              if (j) {
                if (insert)
                  shift_string(cur_note, cursor,
                    cursor < 50 ? 10*(cursor/10)+9 : 58, 1, ch[0]);
                else {
                  ndata[cur_note].text[cursor] = ch[0];
                  print_letter(cur_note, cursor);
                }
                set_cursor(cursor < 58 ? cursor+1 : cursor);
              }
          }
      }
    }
    usleep(mode == M_EDIT ? 10000L : 100000L);  /* sleep 0.1/1 ms */
#if TIMEOUT > 0
    if (mode == M_EDIT && time(0)-idle > TIMEOUT && !opts.no_timeout) {
      set_mask(1);
      if (string_empty(ndata[cur_note].text, 0))
        remove_note(cur_note);
      render_pinboard(-1);
      copy_image_area(board_img, board_img, 0, 0, 64, 0, 64, 64);
      update_display(board_img, 0, 0);
      notes_io(1);
      set_kbfocus(0);
      mode = M_NOOP;
    }
#endif
  }

/* never reached...
  XFreePixmap(display, app.pixmap);
  XFreePixmap(display, app.mask);
  XFreePixmap(display, pixmask);
  XDestroyImage(char_img);
  XDestroyImage(board_img);
  XDestroyImage(note_img);
  XDestroyImage(palette);
  XCloseDisplay(display);
*/
}

